/*
 *  Copyleft (C) 2000 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include "Midi.h"
#include "NoteTable.h"
#include "unistd.h"
#include "sys/types.h"
#include "signal.h"

#ifdef KEYBOARD_SUPPORT
#include <FL/Fl.h>
#endif

static const int MIDI_SCANBUFSIZE=256;
static const int MIDI_KEYOFFSET=0;

static int NKEYS = 30;
static char KEYMAP[30]={'z','s','x','d','c','v','g','b','h','n','j','m','q',
						'2','w','3','e','r','5','t','6','y','7','u','i','9',
						'o','0','p','['};
							
MidiDevice *MidiDevice::m_Singleton;
string MidiDevice::m_DeviceName;

MidiDevice::MidiDevice() :
m_Poly(1)
{
	Open();

#ifdef KEYBOARD_SUPPORT
	m_Oct=4;
	m_CurrentVoice=0;
	
	for (int n=0; n<256; n++)
	{	
		m_KeyVoice[n]=' ';
	}
#endif
}

MidiDevice::~MidiDevice() 
{
	Close();
}

void MidiDevice::Close()
{
	cerr<<"Ending Spiral Midi services"<<endl;

	//if (!SpiralInfo::WANTMIDI) return;
	
	close(m_MidiFd);
	kill(m_ChildPID,9);
}


void MidiDevice::Open()
{
	cerr<<"Starting Spiral Midi services"<<endl;

	//if (!SpiralInfo::WANTMIDI) return;
	
	m_MidiFd = open(m_DeviceName.c_str(),O_RDONLY|O_SYNC);  

	if (!m_MidiFd) 
	{
		cerr<<"Couldn't open midi ["<<m_DeviceName<<"]"<<endl;
		return;
	}
	cerr<<"Opened midi device ["<<m_DeviceName<<"]"<<endl;

	pipe(m_Pipefd);
	fcntl(m_Pipefd[0],F_SETFL,O_NONBLOCK);
	
	m_ChildPID=fork();
	
	if (!m_ChildPID)
	{
		CollectEvents();
	}
}

void MidiDevice::CollectEvents()
{		
	unsigned char buf[1];	
	int count,n,nn;
	bool MessageSent;
	unsigned char data[3],last=0;
	
	// constantly scan for relevent input,
	// and write it to the pipe

	// filters out unhandled messages, and attempts to build 
	// coherent messages to send to the midi handler
		
	for(;;)
	{
		read(m_MidiFd,buf,1);
		
		if (*buf!=248) // odd code that gets sent occasionally
		{
			if (*buf>=0x80) // we've got a opcode
			{
				last=data[0]=*buf;
							
				// find out what the opcode means									
				if(data[0]>=0x90 && data[0]<=0x9f) // note on event
				{	
					// get the next two bytes
					read(m_MidiFd,&data[1],1);
					read(m_MidiFd,&data[2],1);
					write(m_Pipefd[1],data,3); 												
				}
				else if(data[0]>=0x80 && data[0]<=0x8f) // note off event
				{
					// get the next two bytes
					read(m_MidiFd,&data[1],1);	
					read(m_MidiFd,&data[2],1);													  
					write(m_Pipefd[1],data,3); 							
				}
				else if(data[0]>=0xe0 && data[0]<=0xef) // pitch wheel event
				{
					// get the next two bytes
					read(m_MidiFd,&data[1],1);
					read(m_MidiFd,&data[2],1);		
					write(m_Pipefd[1],data,3); 							
				}
				else if(data[0]>=0xb0 && data[0]<=0xbf) // parameter event
				{
					// get the next two byte
					read(m_MidiFd,&data[1],1);
					read(m_MidiFd,&data[2],1);						
					write(m_Pipefd[1],data,3); 							
				}
				else
				{
					last=0; // we don't know this opcode
				}
			}
			else // more data 
			{
				if (last!=0) 
				{
					data[0]=last;
					data[1]=*buf;
					
					// get the last byte if it's not a pressure message
					if(data[0]>=0xd0 && data[0]<=0xdf) 
					{
						data[2]=0;
					}					
					else
					{
						read(m_MidiFd,&data[2],1);
					}
							  
					write(m_Pipefd[1],data,3); 							
				}
			}
		}
	}
}
	
	
void MidiDevice::Check()
{
	unsigned char midi[3];
				
	int count=read(m_Pipefd[0],midi,3);
	
	// no message
	if (count<=0) return;
	
	MidiEvent::type MessageType=MidiEvent::NONE;
	int Volume=0,Note=0,EventDevice=0;

	if (midi[0] >= 0x80 && midi[0] <= 0x8f) // note off
	{	
		MessageType=MidiEvent::OFF;
		Note=midi[1]-MIDI_KEYOFFSET;
		EventDevice=midi[0]-0x80;
	}
	else if (midi[0] >= 0x90 && midi[0] <= 0x9f) // note on
	{
		Volume = midi[2];
		
		// cope with Roland equipment, where note on's 
		// with zero velocity are sent as note offs.
		if (Volume) MessageType=MidiEvent::ON;
		else MessageType=MidiEvent::OFF;

		Note=midi[1]-MIDI_KEYOFFSET;
		EventDevice=midi[0]-0x90;
	}
	else if (midi[0] >= 0xb0 && midi[0] <= 0xbf) // parameter
	{	
		MessageType=MidiEvent::PARAMETER;
		Note=midi[1];
		Volume=midi[2];
		EventDevice=midi[0]-0xb0;
	}
	else if (midi[0] >= 0xe0 && midi[0] <= 0xef) // note pitchbend
	{			
		MessageType=MidiEvent::PITCHBEND;
		// should probably take the first byte into account too.
		Volume=midi[2];			
		EventDevice=midi[0]-0xe0;
	}

	if (EventDevice<0 || EventDevice>15)
	{
		cerr<<"Error - Midi device "<<EventDevice<<" ??"<<endl;
		return;
	}

	m_EventVec[EventDevice].push(MidiEvent(MessageType,Note,Volume));
}

#ifdef KEYBOARD_SUPPORT
void MidiDevice::CheckKeyboard()
{
	Fl::check();
	MidiEvent::type MessageType=MidiEvent::NONE;
	int Volume=0,Note=0,EventDevice=0;

	if (Fl::event_key(FL_F+1)) m_Oct=0; 
	if (Fl::event_key(FL_F+2)) m_Oct=1; 
	if (Fl::event_key(FL_F+3)) m_Oct=2; 
	if (Fl::event_key(FL_F+4)) m_Oct=3; 
	if (Fl::event_key(FL_F+5)) m_Oct=4; 
	if (Fl::event_key(FL_F+6)) m_Oct=5; 
	if (Fl::event_key(FL_F+7)) m_Oct=6; 
	if (Fl::event_key(FL_F+8)) m_Oct=7; 
	if (Fl::event_key(FL_F+9)) m_Oct=8; 
	if (Fl::event_key(FL_F+10)) m_Oct=9; 
	if (Fl::event_key(FL_F+11)) m_Oct=10; 

	int  note=0;
	char KeyChar=0;
	bool KeyPressed=false;

	for (int key=0; key<NKEYS; key++)
	{	
		KeyChar=KEYMAP[key];
		
		// check if a key's been pressed
		if (Fl::event_key(KeyChar)) 
		{
			KeyPressed=true;
	
			// check it was pressed last time
			bool Found=false;
			for (int n=0; n<m_Poly; n++)
			{
				if (m_KeyVoice[n]==KeyChar)
				{
					Found=true;
				}				
			}
			if (!Found)
			{
				Volume = 127;
				MessageType=MidiEvent::ON;
				Note=(m_Oct*12)+note;
				
				// Move to the next voice
				m_CurrentVoice++;
	
				if (m_CurrentVoice>=m_Poly)
				{
					m_CurrentVoice=0;
				}
				
				// this should be the current voice we are using
				m_KeyVoice[m_CurrentVoice]=KeyChar;
			}
		}
		else // it's not pressed down 
		{
			//see if the note was pressed down last time			
			for (int n=0; n<m_Poly; n++)
			{
				if (m_KeyVoice[n]==KeyChar)
				{				
					MessageType=MidiEvent::OFF;
					Note=(m_Oct*12)+note;
					m_KeyVoice[n]=' ';
				}				
			}
		}
		note++;
	}
	
	if (MessageType!=MidiEvent::NONE)
	{
		m_EventVec[EventDevice].push(MidiEvent(MessageType,Note,Volume));
	}
}
#endif

MidiEvent MidiDevice::GetEvent(int Device)
{
	if (Device<0 || Device>15)
	{
		cerr<<"GetEvent: Invalid Midi device "<<Device<<endl;
		return MidiEvent(MidiEvent::NONE,0,0);
	}

	Check();

#ifdef KEYBOARD_SUPPORT
	CheckKeyboard();
#endif

	if (m_EventVec[Device].size()==0)
	{
		return MidiEvent(MidiEvent::NONE,0,0);
	}

	MidiEvent event(m_EventVec[Device].front());
	m_EventVec[Device].pop();
	return event;
}
