/*  SpiralSound
 *  Copyleft (C) 2001 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 
#include "EnvelopePlugin.h"
#include "EnvelopePluginGUI.h"
#include <FL/Fl_Button.h>
#include "SpiralIcon.xpm"

SpiralPlugin* CreateInstance()
{
	return new EnvelopePlugin;
}

Fl_Pixmap* GetIcon()
{
	return new Fl_Pixmap(SpiralIcon_xpm);
}

int GetID()
{
	return 0x0005;
}

///////////////////////////////////////////////////////

EnvelopePlugin::EnvelopePlugin() :
m_Trigger(false),
m_t(-1.0f),
m_Attack(0.0f),
m_Decay(0.5f),
m_Sustain(1.0f),
m_Release(1.0f),
m_Volume(0.5f),
m_TrigThresh(0.01)
{
	m_PluginInfo.Name="Envelope";
	m_PluginInfo.Width=142;
	m_PluginInfo.Height=110;
	m_PluginInfo.NumInputs=2;
	m_PluginInfo.NumOutputs=2;
	m_PluginInfo.PortTips.push_back("Trigger CV");	
	m_PluginInfo.PortTips.push_back("Input");	
	m_PluginInfo.PortTips.push_back("CV");	
	m_PluginInfo.PortTips.push_back("Output");
}

EnvelopePlugin::~EnvelopePlugin()
{
}

PluginInfo &EnvelopePlugin::Initialise(const HostInfo *Host)
{	
	PluginInfo& Info = SpiralPlugin::Initialise(Host);
	Sample *NewSample = new Sample(Host->BUFSIZE);
	m_Output[0]=NewSample;	
	NewSample = new Sample(Host->BUFSIZE);
	m_Output[1]=NewSample;	
	
	m_SampleTime=1.0/(float)(m_HostInfo->SAMPLERATE);
	
	return Info;
}

Fl_Window *EnvelopePlugin::CreateGUI()
{
	m_GUI = new EnvelopePluginGUI(m_PluginInfo.Width,
								  	    m_PluginInfo.Height,
										this,m_HostInfo);
	m_GUI->hide();
	return m_GUI;
}

void EnvelopePlugin::Execute()
{
	Fl::check();	

	float temp=0;
	float Vol=0;
	bool Freeze=false;
	
	// Early out?	
	/*if (m_t<0 && (!m_Input[0] || m_Input[0]->IsEmpty()))
	{
		m_Output[0]->Zero();
		m_Output[1]->Zero();				
		return;
	}*/
		
	for (int n=0; n<m_HostInfo->BUFSIZE; n++)
	{	
		// Check the trigger CV values
		if (GetInputCV(0,n)>m_TrigThresh)
		{
			if (m_Trigger==false) 
			{
				m_t=0.0f; 
				m_Trigger=true;
			}
		}
		else
		{	
			m_Trigger=false;
		}
		
		// if we are in the envelope...
		if (m_t>=0 && m_t<m_Attack+m_Decay+m_Release) 
		{				
			// find out what part of the envelope we are in	
			// in the attack
			if (m_t<m_Attack)
			{			
				// get normalised position to
				// get the volume between 0 and 1
				temp=m_t/m_Attack;
			}
			else
			// in the decay
			if (m_t<m_Attack+m_Decay)
			{
				//if (n==0) cerr<<"in decay"<<endl;
				// normalised position in m_Attack->m_Decay range
				float nt=(m_t-m_Attack)/m_Decay;
				
				// volume between 1 and m_Sustain
				temp=(1-nt)+(m_Sustain*nt);
			}
			else
			// in the release
			{
				//if (n==0) cerr<<"in release"<<endl;
				// normalised position in m_Decay->m_Release range
				float nt=(m_t-(m_Attack+m_Decay))/m_Release;
				
				// volume between m_Sustain and 0			
				temp=m_Sustain*(1-nt);
				
				if (m_Release<0.2f)
				{
					temp=m_Sustain;
				}	
				
				if (m_Trigger) Freeze=true;
			}
			
			temp*=m_Volume;
			
			m_Output[0]->Set(n,(unsigned short)(temp*MAX_LEV));
			m_Output[1]->Set(n,(short)(GetInputVal(1,n)*temp)); 
				
			if (!Freeze) m_t+=m_SampleTime;
		}
		else
		{
			m_Output[0]->Set(n,0);
			m_Output[1]->Set(n,0); 
			
			// if we've run off the end
			if (m_t>m_Attack+m_Decay+m_Release)
			{
				m_t=-1;
				//m_Output[0]->Zero();
				//m_Output[1]->Zero();
				return;
			}

		}
	} 	
}

void EnvelopePlugin::StreamOut(ostream &s)
{
	s<<m_Version<<" "<<m_Attack<<" "<<m_Decay<<" "<<m_Sustain<<" "<<
	m_Release<<" "<<m_Volume<<" "<<m_TrigThresh;
}

void EnvelopePlugin::StreamIn(istream &s)
{
	int version;
	s>>version;
	s>>m_Attack>>m_Decay>>m_Sustain>>m_Release>>m_Volume>>m_TrigThresh;
}

 
int main()
{
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	HostInfo host;
	host.BUFSIZE=256;
	host.SAMPLERATE=44100;
	host.OUTPUTFILE="/dev/dsp";
	host.MIDIFILE="/dev/midi";
	host.POLY=1;
	host.GUI_COLOUR=100;
	
	
	SpiralPlugin* test = CreateInstance();
	test->Initialise(&host);
    test->CreateGUI();
		
    for (;;) 
	{
    	if (!Fl::check()) break;
		test->Execute();    	
  	}
	
	delete test;	
	
	return 1;	
}
