/*  SpiralSound
 *  Copyleft (C) 2001 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include "MidiPlugin.h"
#include "MidiPluginGUI.h"
#include <FL/Fl_Button.h>
#include "../../NoteTable.h"
#include "../../Midi.h"
#include "SpiralIcon.xpm"

int MidiPlugin::m_RefCount = 0;

SpiralPlugin* CreateInstance()
{
	return new MidiPlugin;
}

Fl_Pixmap* GetIcon()
{
	return new Fl_Pixmap(SpiralIcon_xpm);
}

int GetID()
{
	return 0x0002;
}

///////////////////////////////////////////////////////

MidiPlugin::MidiPlugin() :
m_DeviceNum(0),
m_NoteLevel(0),
m_TriggerLevel(0),
m_PitchBendLevel(0),
m_ParameterLevel(0),
m_NoteCut(false)
{
	m_RefCount++;

	m_PluginInfo.Name="Midi";
	m_PluginInfo.Width=85;
	m_PluginInfo.Height=110;
	m_PluginInfo.NumInputs=0;
	m_PluginInfo.NumOutputs=4;
	m_PluginInfo.PortTips.push_back("Note CV");
	m_PluginInfo.PortTips.push_back("Trigger CV");
	m_PluginInfo.PortTips.push_back("PitchBend CV");
	m_PluginInfo.PortTips.push_back("Controller 1 CV");
}

MidiPlugin::~MidiPlugin()
{
	m_RefCount--;
	if (m_RefCount==0) MidiDevice::PackUpAndGoHome();
}

PluginInfo &MidiPlugin::Initialise(const HostInfo *Host)
{	
	PluginInfo& Info= SpiralPlugin::Initialise(Host);
	
	m_Output[0]=new Sample(m_HostInfo->BUFSIZE);
	m_Output[1]=new Sample(m_HostInfo->BUFSIZE);
	m_Output[2]=new Sample(m_HostInfo->BUFSIZE);	
	m_Output[3]=new Sample(m_HostInfo->BUFSIZE);
		
	MidiDevice::SetDeviceName(Host->MIDIFILE);
		
	return Info;
}

Fl_Window *MidiPlugin::CreateGUI()
{
	m_GUI = new MidiPluginGUI(m_PluginInfo.Width,
										  m_PluginInfo.Height,
										  this,m_HostInfo);
	m_GUI->show();
	m_GUI->hide();
	return m_GUI;
}

void MidiPlugin::Execute()
{
 	Fl::check();
	
	// Done to clear IsEmpty field... 
	m_Output[0]->Zero();
	m_Output[1]->Zero();
	m_Output[2]->Zero();
	m_Output[3]->Zero();
	
	MidiEvent Event=MidiDevice::Get()->GetEvent(m_DeviceNum);
	// get all the midi events since the last check
	while(Event.GetType()!=MidiEvent::NONE)
	{
		if (Event.GetType()==MidiEvent::ON)
		{					
			m_NoteLevel=(short)(NoteTable[Event.GetNote()]*FREQ_CONV)-MAX_LEV;			
			m_TriggerLevel=(short)(Event.GetVolume()*CV_CONV);
		}
			
		if (Event.GetType()==MidiEvent::OFF)
		{
			m_TriggerLevel=0;
			if (m_NoteCut) m_NoteLevel=0;
		}
		
		if (Event.GetType()==MidiEvent::PITCHBEND)
		{
			m_PitchBendLevel=(short)(Event.GetVolume()*(CV_CONV*2))+MAX_LEV;	
		}
		
		if (Event.GetType()==MidiEvent::PARAMETER)
		{
			m_ParameterLevel=(unsigned short)(Event.GetNote()+Event.GetVolume()*CV_CONV);			
		}
		
		Event=MidiDevice::Get()->GetEvent(m_DeviceNum);
	}
	for (int n=0; n<m_HostInfo->BUFSIZE; n++)
	{
		m_Output[0]->Set(n,m_NoteLevel);
		m_Output[1]->Set(n,m_TriggerLevel);
		m_Output[2]->Set(n,m_PitchBendLevel);
		m_Output[3]->Set(n,m_ParameterLevel);	
	}
}
void MidiPlugin::StreamOut(ostream &s) 
{
	s<<m_Version<<" "<<m_DeviceNum<<" "<<m_NoteCut<<" ";
}

void MidiPlugin::StreamIn(istream &s) 
{
	int version;
	s>>version;
	s>>m_DeviceNum>>m_NoteCut;
}

int main()
{
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	HostInfo host;
	host.BUFSIZE=256;
	host.SAMPLERATE=44100;
	host.OUTPUTFILE="/dev/dsp";
	host.MIDIFILE="/dev/midi";
	host.POLY=1;
	host.GUI_COLOUR=100;
	
	
	SpiralPlugin* test = CreateInstance();
	test->Initialise(&host);
    Fl_Window* t=test->CreateGUI();
	t->show();
	
    for (;;) 
	{
    	if (!Fl::check()) break;
		test->Execute();    	
  	}
	
	delete test;	
	
	return 1;	
}
