/*  SpiralSound
 *  Copyleft (C) 2001 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 
#include "OscillatorPlugin.h"
#include "OscillatorPluginGUI.h"
#include <FL/Fl_Button.h>
#include "SpiralIcon.xpm"

static const int IN_FREQ  = 0;
static const int IN_PW    = 1;
static const int IN_SHLEN = 2;

static const int OUT_MAIN = 0;

SpiralPlugin* CreateInstance()
{
	return new OscillatorPlugin;
}

Fl_Pixmap* GetIcon()
{
	return new Fl_Pixmap(SpiralIcon_xpm);
}

int GetID()
{
	return 0x0004;
}

///////////////////////////////////////////////////////

OscillatorPlugin::OscillatorPlugin() :
m_Type(SQUARE),
m_Octave(0),
m_FineFreq(1.0f),
m_PulseWidth(0.5f),
m_ModAmount(1.0f),
m_FreqModBuf(NULL),
m_PulseWidthModBuf(NULL),
m_SHModBuf(NULL)
{

	m_CyclePos=0;	
	m_Note=0;
	m_LastFreq=0;
	
	m_PluginInfo.Name="Oscillator";
	m_PluginInfo.Width=245;
	m_PluginInfo.Height=110;
	m_PluginInfo.NumInputs=3;
	m_PluginInfo.NumOutputs=1;
	m_PluginInfo.PortTips.push_back("Frequency CV");
	m_PluginInfo.PortTips.push_back("PulseWidth CV");
	m_PluginInfo.PortTips.push_back("Sample & Hold length CV");
	m_PluginInfo.PortTips.push_back("Output");
}

OscillatorPlugin::~OscillatorPlugin()
{
}

PluginInfo &OscillatorPlugin::Initialise(const HostInfo *Host)
{	
	PluginInfo& Info= SpiralPlugin::Initialise(Host);
	Sample *NewSample = new Sample(Host->BUFSIZE);
	m_Output[OUT_MAIN]=NewSample;
	return Info;
}

Fl_Window *OscillatorPlugin::CreateGUI()
{
	m_GUI = new OscillatorPluginGUI(m_PluginInfo.Width,
										  m_PluginInfo.Height,
										  this,m_HostInfo);									  
  	m_GUI->hide();
	return m_GUI;
}

void OscillatorPlugin::Execute()
{
 	Fl::check();

	static short count = 0, noisev;
	float Freq=0;
	float CycleLen=0;
	int samplelen, PW, InputFreq;
		
	switch (m_Type)
	{
	case SQUARE:
		for (int n=0; n<m_HostInfo->BUFSIZE; n++)
		{
			m_Output[OUT_MAIN]->Set(n,0);		
			if (m_Input[0]) Freq=(unsigned short)(GetInputVal(0,n)+MAX_LEV)/FREQ_CONV;
			else Freq=110;		
			Freq*=m_FineFreq;	
			if (m_Octave>0) Freq*=1<<(m_Octave);
			if (m_Octave<0) Freq/=1<<(-m_Octave);
			CycleLen = m_HostInfo->SAMPLERATE/Freq;
			PW = (int)((m_PulseWidth+GetInputCV(IN_PW,n)*m_ModAmount) * CycleLen);
 	
			// calculate square wave pattern
			m_CyclePos++;
			if (m_CyclePos>CycleLen) m_CyclePos=0;	
				
			if (m_CyclePos<PW) m_Output[OUT_MAIN]->Set(n,MAX_LEV);
			else m_Output[OUT_MAIN]->Set(n,-MAX_LEV);							
		}
		break;
			
	case SAW:
		for (int n=0; n<m_HostInfo->BUFSIZE; n++)
		{
			m_Output[OUT_MAIN]->Set(n,0);		
			if (m_Input[0]) Freq=(unsigned short)(GetInputVal(0,n)+MAX_LEV)/FREQ_CONV;
			else Freq=110;		
			Freq*=m_FineFreq;
			if (m_Octave>0) Freq*=1<<(m_Octave);
			if (m_Octave<0) Freq/=1<<(-m_Octave);			
			CycleLen = m_HostInfo->SAMPLERATE/Freq;
			PW = (int)((m_PulseWidth+GetInputCV(IN_PW,n)*m_ModAmount) * CycleLen);

			// get normailise position between cycle 			
			m_CyclePos++;
			if (m_CyclePos>CycleLen) m_CyclePos=0;		
			
			if (m_CyclePos<PW) 
			{
				// before pw -1->1
				m_Output[OUT_MAIN]->Set(n,Linear(0,PW,m_CyclePos,-MAX_LEV,MAX_LEV));
			}
			else 
			{
				// after pw 1->-1
				m_Output[OUT_MAIN]->Set(n,Linear(PW,CycleLen,m_CyclePos,MAX_LEV,-MAX_LEV));
			}		
		}					
		break;
		
	case NOISE:
		for (int n=0; n<m_HostInfo->BUFSIZE; n++)
		{
			m_CyclePos++;

			//modulate the sample & hold length
			samplelen = (int)((m_SHLen+GetInputCV(IN_SHLEN,n)*m_ModAmount)*m_HostInfo->SAMPLERATE);
		
			// do sample & hold on the noise
			if (m_CyclePos>samplelen)
			{
				noisev=(short)((rand()%MAX_LEV*2)-MAX_LEV);
				m_CyclePos=0;
			}
			m_Output[OUT_MAIN]->Set(n,noisev);
		}
		break;
		
	case NONE: break;
	}
}

void OscillatorPlugin::StreamOut(ostream &s)
{
	s<<m_Version<<" "<<*this;
}

void OscillatorPlugin::StreamIn(istream &s)
{
	int version;
	s>>version>>*this;
}

istream &operator>>(istream &s, OscillatorPlugin &o)
{
	float dummy=0;
	s>>(int&)o.m_Type>>o.m_Octave>>o.m_FineFreq>>o.m_PulseWidth>>dummy>>
	o.m_SHLen>>o.m_ModAmount;
	return s;
}

ostream &operator<<(ostream &s, OscillatorPlugin &o)
{
	float dummy=0;
	s<<(int)o.m_Type<<" "<<o.m_Octave<<" "<<o.m_FineFreq<<" "<<o.m_PulseWidth<<" "<<
	dummy<<" "<<o.m_SHLen<<" "<<o.m_ModAmount<<" ";
	return s;
}

int main()
{
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	HostInfo host;
	host.BUFSIZE=256;
	host.SAMPLERATE=44100;
	host.OUTPUTFILE="/dev/dsp";
	host.MIDIFILE="/dev/midi";
	host.POLY=1;
	host.GUI_COLOUR=100;
	
	
	SpiralPlugin* test = CreateInstance();
	test->Initialise(&host);
    test->CreateGUI();
		
    for (;;) 
	{
    	if (!Fl::check()) break;
		test->Execute();    	
  	}
	
	delete test;	
	
	return 1;	
}
