/*  SpiralSound
 *  Copyleft (C) 2001 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 

#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#if defined (__FreeBSD__)
	#include <machine/soundcard.h>
#else
#if defined (__NetBSD__) || defined (__OpenBSD__)
	#include <soundcard.h>            /* OSS emulation */
#undef ioctl
#else 	                              /* BSDI, Linux, Solaris */
	#include <sys/soundcard.h>
#endif                                /* __NetBSD__ or __OpenBSD__ */
#endif                                /* __FreeBSD__ */
#include <sys/ioctl.h>
#include <limits.h>

#include "OutputPlugin.h"
#include "OutputPluginGUI.h"
#include <FL/Fl_Button.h>
#include "SpiralIcon.xpm"

static const int IN_FREQ  = 0;
static const int IN_PW    = 1;
static const int IN_SHLEN = 2;

static const int OUT_MAIN = 0;

static const HostInfo* host;
OSSOutput* OSSOutput::m_Singleton = NULL;
int OutputPlugin::m_RefCount=0;
int OutputPlugin::m_NoExecuted=0;


SpiralPlugin* CreateInstance()
{
	return new OutputPlugin;
}

Fl_Pixmap* GetIcon()
{	
	return new Fl_Pixmap(SpiralIcon_xpm);
}

int GetID()
{
	return 0x0000;
}

///////////////////////////////////////////////////////

OutputPlugin::OutputPlugin() 
{
	m_RefCount++;
	
	m_PluginInfo.Name="OSS";
	m_PluginInfo.Width=85;
	m_PluginInfo.Height=110;
	m_PluginInfo.NumInputs=2;
	m_PluginInfo.NumOutputs=0;
	m_PluginInfo.PortTips.push_back("Left");
	m_PluginInfo.PortTips.push_back("Right");
}

OutputPlugin::~OutputPlugin()
{
	m_RefCount--;
	if (m_RefCount==0)
	{
		OSSOutput::PackUpAndGoHome();
	}
}

PluginInfo &OutputPlugin::Initialise(const HostInfo *Host)
{	
	PluginInfo& Info= SpiralPlugin::Initialise(Host);
	host=Host;
	OSSOutput::Get()->AllocateBuffer();
	return Info;
}

Fl_Window *OutputPlugin::CreateGUI()
{
	m_GUI = new OutputPluginGUI(m_PluginInfo.Width,
										  m_PluginInfo.Height,
										  this,m_HostInfo);
	m_GUI->hide();
	return m_GUI;
}

void OutputPlugin::Execute()
{
 	Fl::check();
	
	OSSOutput::Get()->SendStereo(GetInput(0),GetInput(1));
	
	// Only Play() once per set of plugins
	m_NoExecuted++;
	if (m_NoExecuted==m_RefCount)
	{	
		OSSOutput::Get()->Play();
		m_NoExecuted=0;
	}
}

////////////////////////////////////////////////////////////////

OSSOutput::OSSOutput() :
m_Amp(0.5),
m_Channels(2),
m_BufferNum(0),
m_OutputOk(true)
{ 
	cerr<<"Opening dsp output"<<endl;
	Init(); 
}

OSSOutput::~OSSOutput()
{
	cerr<<"Closing dsp output"<<endl;
	close(m_Dspfd);
}

void OSSOutput::AllocateBuffer()
{
	// initialise for stereo
	m_Buffer[0].Allocate(host->BUFSIZE*m_Channels);
	m_Buffer[0].Zero();
	m_Buffer[1].Allocate(host->BUFSIZE*m_Channels);
	m_Buffer[1].Zero();
	
	m_Wav.SetSamplerate(host->SAMPLERATE);
}
	
void OSSOutput::SendStereo(const Sample *ldata,const Sample *rdata) 
{
	if (m_Channels!=2) return;
	
	int on=0;
	long temp;
	for (int n=0; n<host->BUFSIZE; n++)
	{
		// stereo channels - interleave	
		if (ldata) m_Buffer[m_BufferNum].Set(on,m_Buffer[m_BufferNum][on]+(*ldata)[n]*m_Amp);
		on++;
		if (rdata) m_Buffer[m_BufferNum].Set(on,m_Buffer[m_BufferNum][on]+(*rdata)[n]*m_Amp);
		on++;
	}
}

void OSSOutput::Play()
{
    int BufferToSend=!m_BufferNum;
	
    #if __BYTE_ORDER == BIG_ENDIAN
    for (int n=0; n<host->BUFSIZE*m_Channels; n++)
    {		
		m_Buffer[BufferToSend].Set(n,(short)((m_Buffer[BufferToSend][n]<<8)&0xff00)|
		                                    ((m_Buffer[BufferToSend][n]>>8)&0xff));
    }
    #endif	
    
	if (m_OutputOk)
		write(m_Dspfd,m_Buffer[BufferToSend].GetBuffer(),sizeof(short)*host->BUFSIZE*m_Channels);
		
    if(m_Wav.Recording()) 
    {
        m_Wav.Save(m_Buffer[BufferToSend]);
    }
	
	m_Buffer[BufferToSend].Zero();
	m_BufferNum=BufferToSend;
}

void OSSOutput::Init()
{ 	
  int r,val;
  
  m_Dspfd = open(host->OUTPUTFILE.c_str(),O_WRONLY);  
  if(m_Dspfd<0) 
  	{
     fprintf(stderr,"Can't open audio driver.\n");
	 m_OutputOk=false;
	 return;
  	}
   
  r = ioctl(m_Dspfd,SNDCTL_DSP_RESET,NULL);

  if (r>=0)
  	{	 
	 short fgmtsize=0;
	 int numfgmts=host->FRAGCOUNT;	 	 
	 if (host->FRAGCOUNT==-1) numfgmts=0x7fff;
	 
	 for (int i=0; i<32; i++)
	 {	 	
	 	if ((host->FRAGSIZE)==(1<<i)) 
		{
			fgmtsize=i; 
			break;
		}
	 }

	 if (fgmtsize==0)
	 {
	 	cerr<<"Fragment size ["<<host->FRAGSIZE<<"] must be power of two!"<<endl;
		exit(0);
	 }
	 	 	
	 numfgmts=numfgmts<<16;
	 val=numfgmts|(int)fgmtsize;
	 r=ioctl(m_Dspfd, SNDCTL_DSP_SETFRAGMENT, &val);
    }
  if(r>=0) 
  	{
	 val = 1;
     r = ioctl(m_Dspfd, SOUND_PCM_WRITE_CHANNELS, &val);
	}
  if(r>=0) 
  	{
     val = AFMT_S16_LE;
     r = ioctl(m_Dspfd,SNDCTL_DSP_SETFMT,&val);	 
  	}
  if(r>=0) 
  	{
	 if (m_Channels==2) val=1;
	 else val=0;
	
     r = ioctl(m_Dspfd,SNDCTL_DSP_STEREO,&val);
  	}
  if(r>=0) 
  	{
     val = host->SAMPLERATE;
     r = ioctl(m_Dspfd,SNDCTL_DSP_SPEED,&val);
  	}
  if(r<0) 
  	{
     fprintf(stderr,"Sound device did not accept settings.\n");
     exit(0);
  	}  
}


int main()
{
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	HostInfo host;
	host.BUFSIZE=256;
	host.SAMPLERATE=44100;
	host.OUTPUTFILE="/dev/dsp";
	host.MIDIFILE="/dev/midi";
	host.POLY=1;
	host.GUI_COLOUR=100;
	
	
	SpiralPlugin* test = CreateInstance();
	test->Initialise(&host);
    test->CreateGUI();
		
    for (;;) 
	{
    	if (!Fl::check()) break;
		test->Execute();    	
  	}
	
	delete test;	
	
	return 1;	
}
