/*  SpiralSound
 *  Copyleft (C) 2001 David Griffiths <dave@pawfal.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/ 
#include "WaveTablePlugin.h"
#include "WaveTablePluginGUI.h"
#include <FL/Fl_Button.h>
#include "SpiralIcon.xpm"

static const int IN_FREQ  = 0;
static const int IN_PW    = 1;
static const int IN_SHLEN = 2;

static const int OUT_MAIN = 0;

SpiralPlugin* CreateInstance()
{
	return new WaveTablePlugin;
}

Fl_Pixmap* GetIcon()
{
	return new Fl_Pixmap(SpiralIcon_xpm);
}

int GetID()
{
	return 0x0017;
}

///////////////////////////////////////////////////////

WaveTablePlugin::WaveTablePlugin() :
m_Type(SINE),
m_Octave(0),
m_FineFreq(1.0f),
m_ModAmount(1.0f),
m_TableLength(DEFAULT_TABLE_LEN)
{

	m_CyclePos=0;	
	m_Note=0;
	
	m_PluginInfo.Name="WaveTable";
	m_PluginInfo.Width=245;
	m_PluginInfo.Height=110;
	m_PluginInfo.NumInputs=1;
	m_PluginInfo.NumOutputs=1;
	m_PluginInfo.PortTips.push_back("Frequency CV");
	m_PluginInfo.PortTips.push_back("Output");
}

WaveTablePlugin::~WaveTablePlugin()
{
}

PluginInfo &WaveTablePlugin::Initialise(const HostInfo *Host)
{	
	PluginInfo& Info= SpiralPlugin::Initialise(Host);
	Sample *NewSample = new Sample(Host->BUFSIZE);
	m_Output[OUT_MAIN]=NewSample;
	
	for (int n=0; n<NUM_TABLES; n++)
	{
		m_Table[n].Allocate(m_TableLength);
	}
	
	WriteWaves();
	
	return Info;
}

Fl_Window *WaveTablePlugin::CreateGUI()
{
	m_GUI = new WaveTablePluginGUI(m_PluginInfo.Width,
										  m_PluginInfo.Height,
										  this,m_HostInfo);									  
  	m_GUI->hide();
	return m_GUI;
}

void WaveTablePlugin::WriteWaves()
{
	float RadCycle = (M_PI/180)*360;
	float Pos=0;

	for (int n=0; n<m_TableLength; n++)
	{
		if (n==0) Pos=0;
		else Pos=(n/(float)m_TableLength)*RadCycle;
		m_Table[SINE].Set(n,sin(Pos)*MAX_LEV);		
	}

	for (int n=0; n<m_TableLength; n++)
	{
		if (n<m_TableLength/2) m_Table[SQUARE].Set(n,MAX_LEV);
		else m_Table[SQUARE].Set(n,-MAX_LEV);				
	}
	
	for (int n=0; n<m_TableLength; n++)
	{
		m_Table[REVSAW].Set(n,((n/(float)m_TableLength)*MAX_LEV*2)-MAX_LEV);
	}
	
	for (int n=0; n<m_TableLength; n++)
	{
		m_Table[SAW].Set(n,MAX_LEV-(n/(float)m_TableLength)*MAX_LEV*2);
	}

	float HalfTab=m_TableLength/2;
	float v=0;
	for (int n=0; n<m_TableLength; n++)
	{
		if (n<HalfTab) v=MAX_LEV-(n/HalfTab)*MAX_LEV*2;
		else v=(((n-HalfTab)/HalfTab)*MAX_LEV*2)-MAX_LEV;
		v*=0.99;
		m_Table[TRIANGLE].Set(n,v);		
	}

	for (int n=0; n<m_TableLength; n++)
	{
		if (n<m_TableLength/1.2) m_Table[PULSE1].Set(n,MAX_LEV);
		else m_Table[PULSE1].Set(n,-MAX_LEV);				
	}
	
	for (int n=0; n<m_TableLength; n++)
	{
		if (n<m_TableLength/1.5) m_Table[PULSE2].Set(n,MAX_LEV);
		else m_Table[PULSE2].Set(n,-MAX_LEV);				
	}

	Pos=0;
	for (int n=0; n<m_TableLength; n++)
	{
		if (n==0) Pos=0;
		else Pos=(n/(float)m_TableLength)*RadCycle;
		if (sin(Pos)==0) m_Table[INVSINE].Set(n,0);
		else m_Table[INVSINE].Set(n,(1.0f/sin(Pos))*MAX_LEV/10.0f);				
	}
}

void WaveTablePlugin::Execute()
{
 	Fl::check();

	static short count = 0, noisev;
	float Freq=0;
	float CycleLen=0;
	int samplelen, PW, InputFreq;

	float Incr;
	
	for (int n=0; n<m_HostInfo->BUFSIZE; n++)
	{	
		m_Output[OUT_MAIN]->Set(n,0);
		
		if (m_Input[0]) 
		{
			Freq=(unsigned short)(GetInputVal(0,n)+MAX_LEV)/FREQ_CONV;
		}
		else 
		{	
			Freq=110;
		}
		
		Freq*=m_FineFreq;
		if (m_Octave>0) Freq*=1<<(m_Octave);
		if (m_Octave<0) Freq/=1<<(-m_Octave);
				
		Incr = Freq*(m_TableLength/(float)m_HostInfo->SAMPLERATE);
		m_CyclePos+=Incr;
		while (m_CyclePos>m_TableLength-1) m_CyclePos-=m_TableLength;
		
		m_Output[OUT_MAIN]->Set(n,m_Table[m_Type][m_CyclePos]);	
	}
}

void WaveTablePlugin::StreamOut(ostream &s)
{
	s<<m_Version<<" "<<*this;
}

void WaveTablePlugin::StreamIn(istream &s)
{
	int version;
	s>>version>>*this;
}

istream &operator>>(istream &s, WaveTablePlugin &o)
{
	s>>(int&)o.m_Type>>o.m_Octave>>o.m_FineFreq>>o.m_ModAmount;
	return s;
}

ostream &operator<<(ostream &s, WaveTablePlugin &o)
{
	s<<(int)o.m_Type<<" "<<o.m_Octave<<" "<<o.m_FineFreq<<" "<<o.m_ModAmount<<" ";
	return s;
}

int main()
{
	Fl::visual(FL_DOUBLE|FL_RGB);
	
	HostInfo host;
	host.BUFSIZE=256;
	host.SAMPLERATE=44100;
	host.OUTPUTFILE="/dev/dsp";
	host.MIDIFILE="/dev/midi";
	host.POLY=1;
	host.GUI_COLOUR=100;
	
	
	SpiralPlugin* test = CreateInstance();
	test->Initialise(&host);
    Fl_Window* t=test->CreateGUI();
	t->show();
	
    for (;;) 
	{
    	if (!Fl::check()) break;
		test->Execute();    	
  	}
	
	delete test;	
	
	return 1;	
}
